@@ -29,12 +29,14 @@
 #include "firebird.h"
 
 #include "../common/classes/ClumpletReader.h"
+#include "../common/classes/MetaName.h"
 #include "fb_exception.h"
 
 #include "../jrd/ibase.h"
+#include "firebird/Interface.h"
 
 #ifdef DEBUG_CLUMPLETS
-#include "../jrd/gds_proto.h"
+#include "../yvalve/gds_proto.h"
 #include <ctype.h>
 
 namespace Firebird {
@@ -42,10 +44,10 @@ namespace Firebird {
 class ClumpletDump : public ClumpletReader
 {
 public:
-	ClumpletDump(Kind k, const UCHAR* buffer, size_t buffLen)
+	ClumpletDump(Kind k, const UCHAR* buffer, FB_SIZE_T buffLen)
 		: ClumpletReader(k, buffer, buffLen)
 	{ }
-	static string hexString(const UCHAR* b, size_t len)
+	static string hexString(const UCHAR* b, FB_SIZE_T len)
 	{
 		string t1, t2;
 		for (; len > 0; --len, ++b)
@@ -96,7 +98,7 @@ void ClumpletReader::dump() const
 	catch (const fatal_exception& x)
 	{
 		gds__log("Fatal exception during clumplet dump: %s", x.what());
-		size_t l = getBufferLength() - getCurOffset();
+		FB_SIZE_T l = getBufferLength() - getCurOffset();
 		const UCHAR *p = getBuffer() + getCurOffset();
 		gds__log("Plain dump starting with offset %d: %s", getCurOffset(),
 			ClumpletDump::hexString(p, l).c_str());
@@ -109,18 +111,74 @@ void ClumpletReader::dump() const
 
 namespace Firebird {
 
-ClumpletReader::ClumpletReader(Kind k, const UCHAR* buffer, size_t buffLen) :
+ClumpletReader::ClumpletReader(Kind k, const UCHAR* buffer, FB_SIZE_T buffLen) :
 	kind(k), static_buffer(buffer), static_buffer_end(buffer + buffLen)
 {
 	rewind();	// this will set cur_offset and spbState
 }
 
-ClumpletReader::ClumpletReader(MemoryPool& pool, Kind k, const UCHAR* buffer, size_t buffLen) :
+ClumpletReader::ClumpletReader(MemoryPool& pool, Kind k, const UCHAR* buffer, FB_SIZE_T buffLen) :
 	AutoStorage(pool), kind(k), static_buffer(buffer), static_buffer_end(buffer + buffLen)
 {
 	rewind();	// this will set cur_offset and spbState
 }
 
+ClumpletReader::ClumpletReader(MemoryPool& pool, const ClumpletReader& from) :
+	AutoStorage(pool), kind(from.kind),
+	static_buffer(from.getBuffer()), static_buffer_end(from.getBufferEnd())
+{
+	rewind();	// this will set cur_offset and spbState
+}
+
+ClumpletReader::ClumpletReader(const ClumpletReader& from) :
+	AutoStorage(), kind(from.kind),
+	static_buffer(from.getBuffer()), static_buffer_end(from.getBufferEnd())
+{
+	rewind();	// this will set cur_offset and spbState
+}
+
+ClumpletReader::ClumpletReader(MemoryPool& pool, const KindList* kl,
+							   const UCHAR* buffer, FB_SIZE_T buffLen, FPTR_VOID raise) :
+	AutoStorage(pool), kind(kl->kind), static_buffer(buffer), static_buffer_end(buffer + buffLen)
+{
+	create(kl, buffLen, raise);
+}
+
+ClumpletReader::ClumpletReader(const KindList* kl, const UCHAR* buffer, FB_SIZE_T buffLen, FPTR_VOID raise) :
+	kind(kl->kind), static_buffer(buffer), static_buffer_end(buffer + buffLen)
+{
+	create(kl, buffLen, raise);
+}
+
+void ClumpletReader::create(const KindList* kl, FB_SIZE_T buffLen, FPTR_VOID raise)
+{
+	cur_offset = 0;
+
+	if (buffLen)
+	{
+		while (kl->kind != EndOfList)
+		{
+			kind = kl->kind;
+			if (getBufferTag() == kl->tag)
+			{
+				break;
+			}
+			++kl;
+		}
+
+		if (kl->kind == EndOfList)
+		{
+			if (raise)
+			{
+				raise();
+			}
+			invalid_structure("Unknown tag value - missing in the list of possible");
+		}
+	}
+
+	rewind();	// this will set cur_offset and spbState
+}
+
 const UCHAR* ClumpletReader::getBuffer() const
 {
 	return static_buffer;
@@ -147,6 +205,20 @@ void ClumpletReader::invalid_structure(const char* what) const
 	fatal_exception::raiseFmt("Invalid clumplet buffer structure: %s", what);
 }
 
+bool ClumpletReader::isTagged() const
+{
+	switch (kind)
+	{
+	case Tpb:
+	case Tagged:
+	case WideTagged:
+	case SpbAttach:
+		return true;
+	}
+
+	return false;
+}
+
 UCHAR ClumpletReader::getBufferTag() const
 {
 	const UCHAR* const buffer_end = getBufferEnd();
@@ -166,7 +238,8 @@ UCHAR ClumpletReader::getBufferTag() const
 	case SpbStart:
 	case UnTagged:
 	case WideUnTagged:
-	case SpbItems:
+	case SpbSendItems:
+	case SpbReceiveItems:
 		usage_mistake("buffer is not tagged");
 		return 0;
 	case SpbAttach:
@@ -189,6 +262,9 @@ UCHAR ClumpletReader::getBufferTag() const
 				return 0;
 			}
 			return buffer_start[1];
+		case isc_spb_version3:
+			// This is wide SPB attach format - buffer's tag is the first byte.
+			return buffer_start[0];
 		default:
 			invalid_structure("spb in service attach should begin with isc_spb_version1 or isc_spb_version");
 			return 0;
@@ -219,9 +295,31 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			return TraditionalDpb;
 		}
 		return SingleTpb;
-	case SpbItems:
+	case SpbSendItems:
+		switch (tag)
+		{
+		case isc_info_svc_auth_block:
+			return Wide;
+		case isc_info_end:
+		case isc_info_truncated:
+		case isc_info_error:
+		case isc_info_data_not_ready:
+		case isc_info_length:
+		case isc_info_flag_end:
+			return SingleTpb;
+		}
+		return StringSpb;
+	case SpbReceiveItems:
 		return SingleTpb;
 	case SpbStart:
+		switch(tag)
+		{
+		case isc_spb_auth_block:
+		case isc_spb_trusted_auth:
+		case isc_spb_auth_plugin_name:
+		case isc_spb_auth_plugin_list:
+			return Wide;
+		}
 		switch (spbState)
 		{
 		case 0:
@@ -235,6 +333,8 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			case isc_spb_res_fix_fss_data:
 			case isc_spb_res_fix_fss_metadata:
 			case isc_spb_bkp_stat:
+			case isc_spb_bkp_skip_data:
+			//case isc_spb_res_skip_data:	// same value
 				return StringSpb;
 			case isc_spb_bkp_factor:
 			case isc_spb_bkp_length:
@@ -242,6 +342,7 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			case isc_spb_res_buffers:
 			case isc_spb_res_page_size:
 			case isc_spb_options:
+			case isc_spb_verbint:
 				return IntSpb;
 			case isc_spb_verbose:
 				return SingleTpb;
@@ -260,6 +361,10 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			case isc_spb_rpr_rollback_trans:
 			case isc_spb_rpr_recover_two_phase:
 				return IntSpb;
+			case isc_spb_rpr_commit_trans_64:
+			case isc_spb_rpr_rollback_trans_64:
+			case isc_spb_rpr_recover_two_phase_64:
+				return BigIntSpb;
 			}
 			invalid_structure("unknown parameter for repair");
 			break;
@@ -320,6 +425,7 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			{
 			case isc_spb_dbname:
 			case isc_spb_command_line:
+			case isc_spb_sts_table:
 				return StringSpb;
 			case isc_spb_options:
 				return IntSpb;
@@ -347,7 +453,7 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 		case isc_action_svc_trace_stop:
 		case isc_action_svc_trace_suspend:
 		case isc_action_svc_trace_resume:
-			switch(tag)
+			switch (tag)
 			{
 			case isc_spb_trc_cfg:
 			case isc_spb_trc_name:
@@ -357,7 +463,7 @@ ClumpletReader::ClumpletType ClumpletReader::getClumpletType(UCHAR tag) const
 			}
 			break;
 		case isc_action_svc_validate:
-			switch(tag)
+			switch (tag)
 			{
 			case isc_spb_val_tab_incl:
 			case isc_spb_val_tab_excl:
@@ -382,7 +488,9 @@ void ClumpletReader::adjustSpbState()
 	switch (kind)
 	{
 	case SpbStart:
-		if (spbState == 0) {	// Just started with service start block
+		if (spbState == 0 &&							// Just started with service start block ...
+			getClumpletSize(true, true, true) == 1)		// and this is action_XXX clumplet
+		{
 			spbState = getClumpTag();
 		}
 		break;
@@ -391,7 +499,7 @@ void ClumpletReader::adjustSpbState()
 	}
 }
 
-size_t ClumpletReader::getClumpletSize(bool wTag, bool wLength, bool wData) const
+FB_SIZE_T ClumpletReader::getClumpletSize(bool wTag, bool wLength, bool wData) const
 {
 	const UCHAR* clumplet = getBuffer() + cur_offset;
 	const UCHAR* const buffer_end = getBufferEnd();
@@ -403,9 +511,9 @@ size_t ClumpletReader::getClumpletSize(bool wTag, bool wLength, bool wData) cons
 		return 0;
 	}
 
-	size_t rc = wTag ? 1 : 0;
-	size_t lengthSize = 0;
-	size_t dataSize = 0;
+	FB_SIZE_T rc = wTag ? 1 : 0;
+	FB_SIZE_T lengthSize = 0;
+	FB_SIZE_T dataSize = 0;
 
 	switch (getClumpletType(clumplet[0]))
 	{
@@ -463,17 +571,22 @@ size_t ClumpletReader::getClumpletSize(bool wTag, bool wLength, bool wData) cons
 		dataSize = 4;
 		break;
 
+	// Used in SPB for 8-byte integers
+	case BigIntSpb:
+		dataSize = 8;
+		break;
+
 	// Used in SPB for single byte
 	case ByteSpb:
 		dataSize = 1;
 		break;
 	}
 
-	const size_t total = 1 + lengthSize + dataSize;
+	const FB_SIZE_T total = 1 + lengthSize + dataSize;
 	if (clumplet + total > buffer_end)
 	{
 		invalid_structure("buffer end before end of clumplet - clumplet too long");
-		size_t delta = total - (buffer_end - clumplet);
+		FB_SIZE_T delta = total - (buffer_end - clumplet);
 		if (delta > dataSize)
 			dataSize = 0;
 		else
@@ -493,7 +606,7 @@ void ClumpletReader::moveNext()
 {
 	if (isEof())
 		return;		// no need to raise useless exceptions
-	size_t cs = getClumpletSize(true, true, true);
+	FB_SIZE_T cs = getClumpletSize(true, true, true);
 	adjustSpbState();
 	cur_offset += cs;
 }
@@ -511,7 +624,8 @@ void ClumpletReader::rewind()
 	case UnTagged:
 	case WideUnTagged:
 	case SpbStart:
-	case SpbItems:
+	case SpbSendItems:
+	case SpbReceiveItems:
 		cur_offset = 0;
 		break;
 	default:
@@ -525,7 +639,7 @@ void ClumpletReader::rewind()
 
 bool ClumpletReader::find(UCHAR tag)
 {
-	const size_t co = getCurOffset();
+	const FB_SIZE_T co = getCurOffset();
 	for (rewind(); !isEof(); moveNext())
 	{
 		if (tag == getClumpTag())
@@ -537,6 +651,27 @@ bool ClumpletReader::find(UCHAR tag)
 	return false;
 }
 
+bool ClumpletReader::next(UCHAR tag)
+{
+	if (!isEof())
+	{
+		const FB_SIZE_T co = getCurOffset();
+		if (tag == getClumpTag())
+		{
+			moveNext();
+		}
+		for (; !isEof(); moveNext())
+		{
+			if (tag == getClumpTag())
+			{
+				return true;
+			}
+		}
+		setCurOffset(co);
+	}
+	return false;
+}
+
 // Methods which work with currently selected clumplet
 UCHAR ClumpletReader::getClumpTag() const
 {
@@ -553,7 +688,7 @@ UCHAR ClumpletReader::getClumpTag() const
 	return clumplet[0];
 }
 
-size_t ClumpletReader::getClumpLength() const
+FB_SIZE_T ClumpletReader::getClumpLength() const
 {
 	return getClumpletSize(false, false, true);
 }
@@ -563,7 +698,7 @@ const UCHAR* ClumpletReader::getBytes() const
 	return getBuffer() + cur_offset + getClumpletSize(true, true, false);
 }
 
-SINT64 ClumpletReader::fromVaxInteger(const UCHAR* ptr, size_t length)
+SINT64 ClumpletReader::fromVaxInteger(const UCHAR* ptr, FB_SIZE_T length)
 {
 	// We can't handle numbers bigger than int64. Some cases use length == 0.
 	fb_assert(ptr && length >= 0 && length < 9);
@@ -582,7 +717,7 @@ SINT64 ClumpletReader::fromVaxInteger(const UCHAR* ptr, size_t length)
 
 SLONG ClumpletReader::getInt() const
 {
-	const size_t length = getClumpLength();
+	const FB_SIZE_T length = getClumpLength();
 
 	if (length > 4)
 	{
@@ -637,7 +772,7 @@ ISC_TIMESTAMP ClumpletReader::getTimeStamp() const
 
 SINT64 ClumpletReader::getBigInt() const
 {
-	const size_t length = getClumpLength();
+	const FB_SIZE_T length = getClumpLength();
 
 	if (length > 8)
 	{
@@ -651,7 +786,7 @@ SINT64 ClumpletReader::getBigInt() const
 string& ClumpletReader::getString(string& str) const
 {
 	const UCHAR* ptr = getBytes();
-	const size_t length = getClumpLength();
+	const FB_SIZE_T length = getClumpLength();
 	str.assign(reinterpret_cast<const char*>(ptr), length);
 	str.recalculate_length();
 	if (str.length() + 1 < length)
@@ -661,10 +796,18 @@ string& ClumpletReader::getString(string& str) const
 	return str;
 }
 
+MetaName& ClumpletReader::getString(MetaName& str) const
+{
+	const UCHAR* ptr = getBytes();
+	const FB_SIZE_T length = getClumpLength();
+	str.assign(reinterpret_cast<const char*>(ptr), length);
+	return str;
+}
+
 PathName& ClumpletReader::getPath(PathName& str) const
 {
 	const UCHAR* ptr = getBytes();
-	const size_t length = getClumpLength();
+	const FB_SIZE_T length = getClumpLength();
 	str.assign(reinterpret_cast<const char*>(ptr), length);
 	str.recalculate_length();
 	if (str.length() + 1 < length)
@@ -674,10 +817,15 @@ PathName& ClumpletReader::getPath(PathName& str) const
 	return str;
 }
 
+void ClumpletReader::getData(UCharBuffer& data) const
+{
+	data.assign(getBytes(), getClumpLength());
+}
+
 bool ClumpletReader::getBoolean() const
 {
 	const UCHAR* ptr = getBytes();
-	const size_t length = getClumpLength();
+	const FB_SIZE_T length = getClumpLength();
 	if (length > 1)
 	{
 		invalid_structure("length of boolean exceeds 1 byte");
@@ -686,5 +834,110 @@ bool ClumpletReader::getBoolean() const
 	return length && ptr[0];
 }
 
-} // namespace
+ClumpletReader::SingleClumplet ClumpletReader::getClumplet() const
+{
+	SingleClumplet rc;
+	rc.tag = getClumpTag();
+	rc.size = getClumpletSize(false, false, true);
+	rc.data = getBytes();
+	return rc;
+}
 
+const ClumpletReader::KindList ClumpletReader::dpbList[] = {
+	{ClumpletReader::Tagged, isc_dpb_version1},
+	{ClumpletReader::WideTagged, isc_dpb_version2},
+	{ClumpletReader::EndOfList, 0}
+};
+
+const ClumpletReader::KindList ClumpletReader::spbList[] = {
+	{ClumpletReader::SpbAttach, isc_spb_current_version},
+	{ClumpletReader::SpbAttach, isc_spb_version1},
+	{ClumpletReader::WideTagged, isc_spb_version3},
+	{ClumpletReader::EndOfList, 0}
+};
+
+AuthReader::AuthReader(const AuthBlock& authBlock)
+	: ClumpletReader(ClumpletReader::WideUnTagged, authBlock.begin(), authBlock.getCount())
+{
+	rewind();
+}
+
+static inline void erase(NoCaseString& s)
+{
+	s.erase();
+}
+
+static inline void set(NoCaseString& s, const ClumpletReader& rdr)
+{
+	s.assign(rdr.getBytes(), rdr.getClumpLength());
+}
+
+bool AuthReader::getInfo(Info& info)
+{
+	if (isEof())
+	{
+		return false;
+	}
+
+	erase(info.type);
+	erase(info.name);
+	erase(info.plugin);
+	erase(info.secDb);
+	erase(info.origPlug);
+
+	ClumpletReader internal(WideUnTagged, getBytes(), getClumpLength());
+	for (internal.rewind(); !internal.isEof(); internal.moveNext())
+	{
+		switch(internal.getClumpTag())
+		{
+		case AUTH_TYPE:
+			set(info.type, internal);
+			break;
+		case AUTH_NAME:
+			set(info.name, internal);
+			break;
+		case AUTH_PLUGIN:
+			set(info.plugin, internal);
+			break;
+		case AUTH_SECURE_DB:
+			set(info.secDb, internal);
+			break;
+		case AUTH_ORIG_PLUG:
+			set(info.origPlug, internal);
+			break;
+		default:
+			break;
+		}
+	}
+
+	return true;
+}
+
+#ifdef AUTH_BLOCK_DEBUG
+void dumpAuthBlock(const char* text, ClumpletReader* pb, unsigned char param)
+{
+	fprintf(stderr, "AuthBlock in %s:", text);
+	if (pb->find(param))
+	{
+		Firebird::AuthReader::AuthBlock tmp;
+		tmp.assign(pb->getBytes(), pb->getClumpLength());
+		Firebird::AuthReader rdr(tmp);
+		string name, plugin;
+		PathName secureDb;
+		bool x = false;
+		while (rdr.getInfo(&name, &plugin, &secureDb))
+		{
+			fprintf(stderr, " %s::%s::%s", name.c_str(), plugin.c_str(), secureDb.c_str());
+			x = true;
+			rdr.moveNext();
+		}
+		fprintf(stderr, "%s\n", x ? "" : " <empty>");
+	}
+	else
+	{
+		fprintf(stderr, " <missing>\n");
+	}
+}
+#endif
+
+} // namespace
