@@ -10,19 +10,20 @@
  *  See the License for the specific language governing rights
  *  and limitations under the License.
  *
- *  The Original Code was created by Vlad Horsun
+ *  The Original Code was created by Vlad Khorsun
  *  for the Firebird Open Source RDBMS project.
  *
- *  Copyright (c) 2005 Vlad Horsun <hvlad@users.sourceforge.net>
+ *  Copyright (c) 2005 Vlad Khorsun <hvlad@users.sourceforge.net>
  *  and all contributors signed below.
  *
  *  All Rights Reserved.
  *  Contributor(s): ______________________________________.
  */
 
+#include "firebird.h"
 #include "../jrd/Relation.h"
-#include "../jrd/tra.h"
 
+#include "../jrd/tra.h"
 #include "../jrd/btr_proto.h"
 #include "../jrd/dpm_proto.h"
 #include "../jrd/idx_proto.h"
@@ -33,98 +34,24 @@
 
 using namespace Jrd;
 
-#ifdef GARBAGE_THREAD
+/// jrd_rel
 
-void RelationGarbage::clear()
-{
-	TranGarbage *item = array.begin(), *const last = array.end();
-
-	for (; item < last; item++)
-	{
-		delete item->bm;
-		item->bm = NULL;
-	}
-
-	array.clear();
-}
-
-void RelationGarbage::addPage(MemoryPool* pool, const SLONG pageno, const SLONG tranid)
-{
-	bool found = false;
-	TranGarbage const *item = array.begin(), *const last = array.end();
-
-	for (; item < last; item++)
-	{
-		if (item->tran <= tranid)
-		{
-			if (PageBitmap::test(item->bm, pageno))
-			{
-				found = true;
-				break;
-			}
-		}
-		else
-		{
-			if (item->bm->clear(pageno))
-				break;
-		}
-	}
-
-	if (!found)
-	{
-		PageBitmap *bm = NULL;
-		size_t pos = 0;
-
-		if (array.find(tranid, pos) )
-		{
-			bm = array[pos].bm;
-			PBM_SET(pool, &bm, pageno);
-		}
-		else
-		{
-			bm = NULL;
-			PBM_SET(pool, &bm, pageno);
-			array.add(TranGarbage(bm, tranid));
-		}
-	}
-}
-
-void RelationGarbage::getGarbage(const SLONG oldest_snapshot, PageBitmap **sbm)
-{
-	while (array.getCount() > 0)
-	{
-		TranGarbage& garbage = array[0];
-
-		if (garbage.tran >= oldest_snapshot)
-			break;
-
-		PageBitmap* bm_tran = garbage.bm;
-		PageBitmap** bm_or = PageBitmap::bit_or(sbm, &bm_tran);
-		if (*bm_or == garbage.bm)
-		{
-			bm_tran = *sbm;
-			*sbm = garbage.bm;
-			garbage.bm = bm_tran;
-		}
-		delete garbage.bm;
-
-		// Need to cast zero to exact type because literal zero means null pointer
-		array.remove(static_cast<size_t>(0));
-	}
-}
-
-#endif //GARBAGE_THREAD
-
-RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool allocPages)
+RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, TraNumber tran, bool allocPages)
 {
 	if (tdbb->tdbb_flags & TDBB_use_db_page_space)
 		return &rel_pages_base;
 
+	Jrd::Attachment* attachment = tdbb->getAttachment();
 	Database* dbb = tdbb->getDatabase();
-	SLONG inst_id;
+
+	SINT64 inst_id;
+	// Vlad asked for this compile-time check to make sure we can contain a txn number here
+	typedef int RangeCheck1[sizeof(inst_id) >= sizeof(TraNumber)];
+	typedef int RangeCheck2[sizeof(inst_id) >= sizeof(AttNumber)];
+
 	if (rel_flags & REL_temp_tran)
 	{
-		if (tran > 0)
+		if (tran > 0 && tran != MAX_TRA_NUMBER) //if (tran > 0)
 			inst_id = tran;
 		else if (tdbb->tdbb_temp_traid)
 			inst_id = tdbb->tdbb_temp_traid;
@@ -137,12 +64,9 @@ RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool alloc
 		inst_id = PAG_attachment_id(tdbb);
 
 	if (!rel_pages_inst)
-	{
-		MemoryPool& pool = *dbb->dbb_permanent;
-		rel_pages_inst = FB_NEW(pool) RelationPagesInstances(pool);
-	}
+		rel_pages_inst = FB_NEW_POOL(*rel_pool) RelationPagesInstances(*rel_pool);
 
-	size_t pos;
+	FB_SIZE_T pos;
 	if (!rel_pages_inst->find(inst_id, pos))
 	{
 		if (!allocPages)
@@ -154,7 +78,7 @@ RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool alloc
 			const size_t BULK_ALLOC = 8;
 
 			RelationPages* allocatedPages = newPages =
-				FB_NEW(*dbb->dbb_permanent) RelationPages[BULK_ALLOC];
+				FB_NEW_POOL(*rel_pool) RelationPages[BULK_ALLOC];
 
 			rel_pages_free = ++allocatedPages;
 			for (size_t i = 1; i < BULK_ALLOC - 1; i++, allocatedPages++)
@@ -177,28 +101,25 @@ RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool alloc
 		DPM_create_relation_pages(tdbb, this, newPages);
 
 #ifdef VIO_DEBUG
-		if (debug_flag > DEBUG_WRITES)
-		{
-			printf("jrd_rel::getPages inst %" SLONGFORMAT", ppp %" SLONGFORMAT", irp %" SLONGFORMAT", addr 0x%x\n",
-				newPages->rel_instance_id,
-				newPages->rel_pages ? (*newPages->rel_pages)[0] : 0,
-				newPages->rel_index_root,
-				newPages);
-		}
+		VIO_trace(DEBUG_WRITES,
+			"jrd_rel::getPages inst %" ULONGFORMAT", ppp %" SLONGFORMAT", irp %" SLONGFORMAT", addr 0x%x\n",
+			newPages->rel_instance_id,
+			newPages->rel_pages ? (*newPages->rel_pages)[0] : 0,
+			newPages->rel_index_root,
+			newPages);
 #endif
 
 		// create indexes
-		MemoryPool *pool = tdbb->getDefaultPool();
+		MemoryPool* pool = tdbb->getDefaultPool();
 		const bool poolCreated = !pool;
 
 		if (poolCreated)
 			pool = dbb->createPool();
 		Jrd::ContextPoolHolder context(tdbb, pool);
 
-		jrd_tra *idx_tran = tdbb->getTransaction();
-		if (!idx_tran) {
-			idx_tran = dbb->dbb_sys_trans;
-		}
+		jrd_tra* idxTran = tdbb->getTransaction();
+		if (!idxTran)
+			idxTran = attachment->getSysTransaction();
 
 		IndexDescAlloc* indices = NULL;
 		// read indices from "base" index root page
@@ -212,18 +133,16 @@ RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool alloc
 
 			idx->idx_root = 0;
 			SelectivityList selectivity(*pool);
-			IDX_create_index(tdbb, this, idx, idx_name.c_str(), NULL, idx_tran, selectivity);
+			IDX_create_index(tdbb, this, idx, idx_name.c_str(), NULL, idxTran, selectivity);
 
 #ifdef VIO_DEBUG
-			if (debug_flag > DEBUG_WRITES)
-			{
-				printf("jrd_rel::getPages inst %" SLONGFORMAT", irp %" SLONGFORMAT", idx %u, idx_root %" SLONGFORMAT", addr 0x%x\n",
-					newPages->rel_instance_id,
-					newPages->rel_index_root,
-					idx->idx_id,
-					idx->idx_root,
-					newPages);
-			}
+			VIO_trace(DEBUG_WRITES,
+				"jrd_rel::getPages inst %" SQUADFORMAT", irp %" SLONGFORMAT", idx %u, idx_root %" SLONGFORMAT", addr 0x%x\n",
+				newPages->rel_instance_id,
+				newPages->rel_index_root,
+				idx->idx_id,
+				idx->idx_root,
+				newPages);
 #endif
 		}
 
@@ -239,13 +158,15 @@ RelationPages* jrd_rel::getPagesInternal(thread_db* tdbb, SLONG tran, bool alloc
 	return pages;
 }
 
-bool jrd_rel::delPages(thread_db* tdbb, SLONG tran, RelationPages* aPages)
+bool jrd_rel::delPages(thread_db* tdbb, TraNumber tran, RelationPages* aPages)
 {
 	RelationPages* pages = aPages ? aPages : getPages(tdbb, tran, false);
 	if (!pages || !pages->rel_instance_id)
 		return false;
 
-	fb_assert((tran <= 0) || ((tran > 0) && (pages->rel_instance_id == tran)));
+	//fb_assert((tran <= 0) || ((tran > 0) && (pages->rel_instance_id == tran)));
+	fb_assert(tran == 0 || tran == MAX_TRA_NUMBER ||
+		(tran > 0 && pages->rel_instance_id == tran));
 
 	fb_assert(pages->useCount > 0);
 
@@ -253,17 +174,15 @@ bool jrd_rel::delPages(thread_db* tdbb, SLONG tran, RelationPages* aPages)
 		return false;
 
 #ifdef VIO_DEBUG
-	if (debug_flag > DEBUG_WRITES)
-	{
-		printf("jrd_rel::delPages inst %" SLONGFORMAT", ppp %" SLONGFORMAT", irp %" SLONGFORMAT", addr 0x%x\n",
-			pages->rel_instance_id,
-			pages->rel_pages ? (*pages->rel_pages)[0] : 0,
-			pages->rel_index_root,
-			pages);
-	}
+	VIO_trace(DEBUG_WRITES,
+		"jrd_rel::delPages inst %" ULONGFORMAT", ppp %" SLONGFORMAT", irp %" SLONGFORMAT", addr 0x%x\n",
+		pages->rel_instance_id,
+		pages->rel_pages ? (*pages->rel_pages)[0] : 0,
+		pages->rel_index_root,
+		pages);
 #endif
 
-	size_t pos;
+	FB_SIZE_T pos;
 #ifdef DEV_BUILD
 	const bool found =
 #endif
@@ -272,13 +191,11 @@ bool jrd_rel::delPages(thread_db* tdbb, SLONG tran, RelationPages* aPages)
 
 	rel_pages_inst->remove(pos);
 
-	if (pages->rel_index_root) {
+	if (pages->rel_index_root)
 		IDX_delete_indices(tdbb, this, pages);
-	}
 
-	if (pages->rel_pages) {
+	if (pages->rel_pages)
 		DPM_delete_relation_pages(tdbb, this, pages);
-	}
 
 	pages->free(rel_pages_free);
 	return true;
@@ -290,13 +207,13 @@ void jrd_rel::getRelLockKey(thread_db* tdbb, UCHAR* key)
 	memcpy(key, &val, sizeof(ULONG));
 	key += sizeof(ULONG);
 
-	const SLONG inst_id = getPages(tdbb)->rel_instance_id;
-	memcpy(key, &inst_id, sizeof(SLONG));
+	const SINT64 inst_id = getPages(tdbb)->rel_instance_id;
+	memcpy(key, &inst_id, sizeof(SINT64));
 }
 
-SSHORT jrd_rel::getRelLockKeyLength() const
+USHORT jrd_rel::getRelLockKeyLength() const
 {
-	return sizeof(ULONG) + sizeof(SLONG);
+	return sizeof(ULONG) + sizeof(SINT64);
 }
 
 void jrd_rel::cleanUp()
@@ -310,7 +227,7 @@ void jrd_rel::fillPagesSnapshot(RelPagesSnapshot& snapshot, const bool attachmen
 {
 	if (rel_pages_inst)
 	{
-		for (size_t i = 0; i < rel_pages_inst->getCount(); i++)
+		for (FB_SIZE_T i = 0; i < rel_pages_inst->getCount(); i++)
 		{
 			RelationPages* relPages = (*rel_pages_inst)[i];
 
@@ -320,7 +237,7 @@ void jrd_rel::fillPagesSnapshot(RelPagesSnapshot& snapshot, const bool attachmen
 				relPages->addRef();
 			}
 			else if ((rel_flags & REL_temp_conn) &&
-				(PAG_attachment_id(snapshot.spt_tdbb) == relPages->rel_instance_id))
+				PAG_attachment_id(snapshot.spt_tdbb) == relPages->rel_instance_id)
 			{
 				snapshot.add(relPages);
 				relPages->addRef();
@@ -351,12 +268,12 @@ void jrd_rel::RelPagesSnapshot::clear()
 	fb_assert(tdbb == spt_tdbb);
 #endif
 
-	for (size_t i = 0; i < getCount(); i++)
+	for (FB_SIZE_T i = 0; i < getCount(); i++)
 	{
 		RelationPages* relPages = (*this)[i];
 		(*this)[i] = NULL;
 
-		spt_relation->delPages(spt_tdbb, -1, relPages);
+		spt_relation->delPages(spt_tdbb, MAX_TRA_NUMBER, relPages);
 	}
 
 	inherited::clear();
@@ -385,34 +302,27 @@ bool jrd_rel::hasTriggers() const
 
 Lock* jrd_rel::createLock(thread_db* tdbb, MemoryPool* pool, jrd_rel* relation, lck_t lckType, bool noAst)
 {
-	Database *dbb = tdbb->getDatabase();
 	if (!pool)
-		pool = dbb->dbb_permanent;
+		pool = relation->rel_pool;
 
-	const SSHORT relLockLen = relation->getRelLockKeyLength();
+	const USHORT relLockLen = relation->getRelLockKeyLength();
 
-	Lock* lock = FB_NEW_RPT(*pool, relLockLen) Lock();
-	lock->lck_dbb = dbb;
-
-	lock->lck_length = relLockLen;
-	relation->getRelLockKey(tdbb, &lock->lck_key.lck_string[0]);
+	Lock* lock = FB_NEW_RPT(*pool, relLockLen) Lock(tdbb, relLockLen, lckType, relation);
+	relation->getRelLockKey(tdbb, lock->getKeyString());
 
 	lock->lck_type = lckType;
 	switch (lckType)
 	{
-	case LCK_relation: 
+	case LCK_relation:
 		break;
 
-	case LCK_rel_gc: 
+	case LCK_rel_gc:
 		lock->lck_ast = noAst ? NULL : blocking_ast_gcLock;
 		break;
 
 	default:
 		fb_assert(false);
 	}
-	lock->lck_owner_handle = LCK_get_owner_handle(tdbb, lock->lck_type);
-	lock->lck_parent = dbb->dbb_lock;
-	lock->lck_object = relation;
 
 	return lock;
 }
@@ -427,7 +337,7 @@ bool jrd_rel::acquireGCLock(thread_db* tdbb, int wait)
 		return true;
 	}
 
-	if (!rel_gc_lock) 
+	if (!rel_gc_lock)
 		rel_gc_lock = createLock(tdbb, NULL, this, LCK_rel_gc, false);
 
 	fb_assert(!rel_gc_lock->lck_id);
@@ -475,11 +385,12 @@ void jrd_rel::downgradeGCLock(thread_db* tdbb)
 
 int jrd_rel::blocking_ast_gcLock(void* ast_object)
 {
-/****
+	/****
 	SR - gc forbidden, awaiting moment to re-establish SW lock
 	SW - gc allowed, usual state
 	PW - gc allowed to the one connection only
-****/
+	****/
+
 	jrd_rel* relation = static_cast<jrd_rel*>(ast_object);
 
 	try
@@ -487,10 +398,10 @@ int jrd_rel::blocking_ast_gcLock(void* ast_object)
 		Lock* lock = relation->rel_gc_lock;
 		Database* dbb = lock->lck_dbb;
 
-		AstContextHolder tdbb(dbb, lock->lck_attachment);
+		AsyncContextHolder tdbb(dbb, FB_FUNCTION, lock);
 
 		fb_assert(!(relation->rel_flags & REL_gc_lockneed));
-		if (relation->rel_flags & REL_gc_lockneed) // work already done syncronously ?
+		if (relation->rel_flags & REL_gc_lockneed) // work already done synchronously ?
 			return 0;
 
 		relation->rel_flags |= REL_gc_blocking;
@@ -508,7 +419,7 @@ int jrd_rel::blocking_ast_gcLock(void* ast_object)
 			relation->rel_flags &= ~(REL_gc_disabled | REL_gc_blocking);
 			relation->rel_flags |= REL_gc_lockneed;
 		}
-		else 
+		else
 		{
 			// someone acquired PW lock
 
@@ -529,7 +440,7 @@ int jrd_rel::blocking_ast_gcLock(void* ast_object)
 /// jrd_rel::GCExclusive
 
 jrd_rel::GCExclusive::GCExclusive(thread_db* tdbb, jrd_rel* relation) :
-	m_tdbb(tdbb), 
+	m_tdbb(tdbb),
 	m_relation(relation),
 	m_lock(NULL)
 {
@@ -554,8 +465,8 @@ bool jrd_rel::GCExclusive::acquire(int wait)
 	int sleeps = -wait * 10;
 	while (m_relation->rel_sweep_count)
 	{
-		Database::Checkout cout(m_tdbb->getDatabase());
-		THD_sleep(100);
+		EngineCheckout cout(m_tdbb, FB_FUNCTION);
+		Thread::sleep(100);
 
 		if (wait < 0 && --sleeps == 0)
 			break;
@@ -615,5 +526,6 @@ void RelationPages::free(RelationPages*& nextFree)
 		rel_pages->clear();
 
 	rel_index_root = rel_data_pages = 0;
-	rel_slot_space = rel_data_space = rel_instance_id = 0;
+	rel_slot_space = rel_pri_data_space = rel_sec_data_space = 0;
+	rel_instance_id = 0;
 }
