@@ -27,17 +27,17 @@
  */
 
 #include "firebird.h"
-#include "../jrd/common.h"
 #include <stdio.h>
 #include "../jrd/jrd.h"
 #include "../jrd/lck.h"
 #include "gen/iberror.h"
 #include "../jrd/err_proto.h"
-#include "../jrd/gds_proto.h"
+#include "../yvalve/gds_proto.h"
 #include "../jrd/jrd_proto.h"
 #include "../jrd/lck_proto.h"
-#include "../jrd/gdsassert.h"
+#include "../common/gdsassert.h"
 #include "../lock/lock_proto.h"
+#include "../jrd/Attachment.h"
 
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
@@ -57,7 +57,7 @@ using namespace Firebird;
 
 static void bug_lck(const TEXT*);
 static bool compatible(const Lock*, const Lock*, USHORT);
-static void enqueue(thread_db*, Lock*, USHORT, SSHORT);
+static void enqueue(thread_db*, CheckStatusWrapper*, Lock*, USHORT, SSHORT);
 static int external_ast(void*);
 static USHORT hash_func(const UCHAR*, USHORT);
 static void hash_allocate(Lock*);
@@ -67,42 +67,45 @@ static bool hash_remove_lock(Lock*, Lock**);
 static void internal_ast(Lock*);
 static bool internal_compatible(Lock*, const Lock*, USHORT);
 static void internal_dequeue(thread_db*, Lock*);
-static USHORT internal_downgrade(thread_db*, Lock*);
-static bool internal_enqueue(thread_db*, Lock*, USHORT, SSHORT, bool);
+static USHORT internal_downgrade(thread_db*, CheckStatusWrapper*, Lock*);
+static bool internal_enqueue(thread_db*, CheckStatusWrapper*, Lock*, USHORT, SSHORT, bool);
+static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type);
 
-static void set_lock_attachment(Lock*, Attachment*);
+#ifdef DEBUG_LCK
+namespace
+{
+	class LckSync
+	{
+	public:
+		LckSync(Lock* lock, const char* sWhere)
+			: m_sync(&lock->lck_sync, sWhere)
+		{
+			/***ThreadSync* thd =***/ ThreadSync::getThread(NULL);
+			m_sync.lock(SYNC_EXCLUSIVE);
+		}
 
+		~LckSync()
+		{
+		}
+
+	private:
+		Sync m_sync;
+	};
+}
+#endif
 
 // globals and macros
 
-#ifdef SUPERSERVER
-
 inline LOCK_OWNER_T LCK_OWNER_ID_DBB(thread_db* tdbb)
 {
 	return (LOCK_OWNER_T) getpid() << 32 | tdbb->getDatabase()->dbb_lock_owner_id;
 }
+
 inline LOCK_OWNER_T LCK_OWNER_ID_ATT(thread_db* tdbb)
 {
-	return (LOCK_OWNER_T) getpid() << 32 | tdbb->getAttachment()->att_lock_owner_id;
-}
+	if (tdbb->getDatabase()->dbb_flags & DBB_shared)
+		return (LOCK_OWNER_T) getpid() << 32 | tdbb->getAttachment()->att_lock_owner_id;
 
-inline SLONG* LCK_OWNER_HANDLE_DBB(thread_db* tdbb)
-{
-	return &tdbb->getDatabase()->dbb_lock_owner_handle;
-}
-inline SLONG* LCK_OWNER_HANDLE_ATT(thread_db* tdbb)
-{
-	return &tdbb->getAttachment()->att_lock_owner_handle;
-}
-
-#else	// SUPERSERVER
-
-inline LOCK_OWNER_T LCK_OWNER_ID_DBB(thread_db* tdbb)
-{
-	return (LOCK_OWNER_T) getpid() << 32 | tdbb->getDatabase()->dbb_lock_owner_id;
-}
-inline LOCK_OWNER_T LCK_OWNER_ID_ATT(thread_db* tdbb)
-{
 	return (LOCK_OWNER_T) getpid() << 32 | tdbb->getDatabase()->dbb_lock_owner_id;
 }
 
@@ -110,13 +113,15 @@ inline SLONG* LCK_OWNER_HANDLE_DBB(thread_db* tdbb)
 {
 	return &tdbb->getDatabase()->dbb_lock_owner_handle;
 }
+
 inline SLONG* LCK_OWNER_HANDLE_ATT(thread_db* tdbb)
 {
+	if (tdbb->getDatabase()->dbb_flags & DBB_shared)
+		return &tdbb->getAttachment()->att_lock_owner_handle;
+
 	return &tdbb->getDatabase()->dbb_lock_owner_handle;
 }
 
-#endif	// SUPERSERVER
-
 
 static const bool compatibility[LCK_max][LCK_max] =
 {
@@ -135,22 +140,22 @@ static const bool compatibility[LCK_max][LCK_max] =
 //#define COMPATIBLE(st1, st2)	compatibility [st1 * LCK_max + st2]
 const int LOCK_HASH_SIZE	= 19;
 
-
-inline void ENQUEUE(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
+inline void ENQUEUE(thread_db* tdbb, CheckStatusWrapper* statusVector, Lock* lock, USHORT level, SSHORT wait)
 {
 	if (lock->lck_compatible)
-		internal_enqueue(tdbb, lock, level, wait, false);
+		internal_enqueue(tdbb, statusVector, lock, level, wait, false);
 	else
-		enqueue(tdbb, lock, level, wait);
+		enqueue(tdbb, statusVector, lock, level, wait);
 }
 
-inline bool CONVERT(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
+inline bool CONVERT(thread_db* tdbb, CheckStatusWrapper* statusVector, Lock* lock, USHORT level, SSHORT wait)
 {
 	Database* const dbb = tdbb->getDatabase();
 
 	return lock->lck_compatible ?
-		internal_enqueue(tdbb, lock, level, wait, true) :
-		dbb->dbb_lock_mgr->convert(tdbb, lock->lck_id, level, wait, lock->lck_ast, lock->lck_object);
+		internal_enqueue(tdbb, statusVector, lock, level, wait, true) :
+		dbb->dbb_lock_mgr->convert(tdbb, statusVector, lock->lck_id, level, wait, lock->lck_ast,
+			lock->lck_object);
 }
 
 inline void DEQUEUE(thread_db* tdbb, Lock* lock)
@@ -167,9 +172,15 @@ inline USHORT DOWNGRADE(thread_db* tdbb, Lock* lock)
 {
 	Database* const dbb = tdbb->getDatabase();
 
-	return lock->lck_compatible ?
-		internal_downgrade(tdbb, lock) :
-		dbb->dbb_lock_mgr->downgrade(tdbb, lock->lck_id);
+	FbLocalStatus statusVector;
+
+	USHORT ret = lock->lck_compatible ?
+		internal_downgrade(tdbb, &statusVector, lock) :
+		dbb->dbb_lock_mgr->downgrade(tdbb, &statusVector, lock->lck_id);
+
+	fb_assert(statusVector.isEmpty());
+
+	return ret;
 }
 
 #ifdef DEV_BUILD
@@ -198,7 +209,7 @@ inline bool checkLock(const Lock* l)
 
 /* The following check should be part of LCK_CHECK_LOCK, but it fails
    when the exclusive attachment to a database is changed to a shared
-   attachment.  When that occurs we fb_assert all our internal locks, but
+   attachment.  When that occurs we assert all our internal locks, but
    while we are in the process of asserting DBB_assert_locks is set, but
    we haven't gotten physical locks yet.
 
@@ -222,15 +233,15 @@ namespace {
 class WaitCancelGuard
 {
 public:
-	WaitCancelGuard(thread_db* tdbb, Lock* lock, int wait) :
-	  m_tdbb(tdbb),
-	  m_save_lock(NULL)
+	WaitCancelGuard(thread_db* tdbb, Lock* lock, int wait)
+		: m_tdbb(tdbb),
+		  m_save_lock(NULL)
 	{
-		Attachment* att = m_tdbb->getAttachment();
+		Jrd::Attachment* att = m_tdbb->getAttachment();
 		if (att)
 			m_save_lock = att->att_wait_lock;
 
-		m_cancel_disabled = (m_tdbb->tdbb_flags & TDBB_wait_cancel_disable); 
+		m_cancel_disabled = (m_tdbb->tdbb_flags & TDBB_wait_cancel_disable);
 		if (wait == LCK_WAIT)
 		{
 			switch (lock->lck_type)
@@ -239,13 +250,13 @@ public:
 				m_tdbb->tdbb_flags &= ~TDBB_wait_cancel_disable;
 				if (att)
 					att->att_wait_lock = lock;
-			break;
+				break;
 
 			default:
 				m_tdbb->tdbb_flags |= TDBB_wait_cancel_disable;
 			}
 		}
-		else
+		else if (wait != LCK_NO_WAIT)
 		{
 			m_tdbb->tdbb_flags &= ~TDBB_wait_cancel_disable;
 			if (att)
@@ -255,11 +266,11 @@ public:
 
 	~WaitCancelGuard()
 	{
-		Attachment* att = m_tdbb->getAttachment();
+		Jrd::Attachment* att = m_tdbb->getAttachment();
 		if (att)
 			att->att_wait_lock = m_save_lock;
 
-		if (m_cancel_disabled) 
+		if (m_cancel_disabled)
 			m_tdbb->tdbb_flags |= TDBB_wait_cancel_disable;
 		else
 			m_tdbb->tdbb_flags &= ~TDBB_wait_cancel_disable;
@@ -316,30 +327,38 @@ bool LCK_convert(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
 	SET_TDBB(tdbb);
 	fb_assert(LCK_CHECK_LOCK(lock));
 
+#ifdef DEBUG_LCK
+	LckSync sync(lock, "LCK_convert");
+#endif
+
 	Database* dbb = lock->lck_dbb;
 
-	Attachment* const old_attachment = lock->lck_attachment;
-	set_lock_attachment(lock, tdbb->getAttachment());
+	Jrd::Attachment* const old_attachment = lock->getLockAttachment();
+	lock->setLockAttachment(tdbb, tdbb->getAttachment());
 
 	WaitCancelGuard guard(tdbb, lock, wait);
-	const bool result = CONVERT(tdbb, lock, level, wait);
+	FbLocalStatus statusVector;
+
+	const bool result = CONVERT(tdbb, &statusVector, lock, level, wait);
 
 	if (!result)
 	{
-	    set_lock_attachment(lock, old_attachment);
+	    lock->setLockAttachment(tdbb, old_attachment);
 
-		switch (tdbb->tdbb_status_vector[1])
+		switch (statusVector[1])
 		{
 		case isc_deadlock:
 		case isc_lock_conflict:
 		case isc_lock_timeout:
+			statusVector.copyTo(tdbb->tdbb_status_vector);
 			tdbb->checkCancelState(true);
 			return false;
 		case isc_lockmanerr:
 			dbb->dbb_flags |= DBB_bugcheck;
 			break;
 		}
-		ERR_punt();
+
+		statusVector.raise();
 	}
 
 	if (!lock->lck_compatible)
@@ -380,7 +399,7 @@ bool LCK_convert_opt(thread_db* tdbb, Lock* lock, USHORT level)
 }
 
 
-bool LCK_cancel_wait(Attachment* attachment)
+bool LCK_cancel_wait(Jrd::Attachment* attachment)
 {
 /**************************************
  *
@@ -389,7 +408,7 @@ bool LCK_cancel_wait(Attachment* attachment)
  **************************************
  *
  * Functional description
- *	Try to cancel waiting of attachment inside the LM	
+ *	Try to cancel waiting of attachment inside the LM.
  *
  **************************************/
 	Database *dbb = attachment->att_database;
@@ -416,6 +435,10 @@ void LCK_downgrade(thread_db* tdbb, Lock* lock)
 	SET_TDBB(tdbb);
 	fb_assert(LCK_CHECK_LOCK(lock));
 
+#ifdef DEBUG_LCK
+	LckSync sync(lock, "LCK_downgrade");
+#endif
+
 	if (lock->lck_id && lock->lck_physical != LCK_none)
 	{
 		const USHORT level = DOWNGRADE(tdbb, lock);
@@ -426,7 +449,7 @@ void LCK_downgrade(thread_db* tdbb, Lock* lock)
 	if (lock->lck_physical == LCK_none)
 	{
 		lock->lck_id = lock->lck_data = 0;
-	    set_lock_attachment(lock, NULL);
+	    lock->setLockAttachment(tdbb, NULL);
 	}
 
 	fb_assert(LCK_CHECK_LOCK(lock));
@@ -469,11 +492,11 @@ void LCK_fini(thread_db* tdbb, enum lck_owner_t owner_type)
 }
 
 
-SLONG LCK_get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
+static SLONG get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 {
 /**************************************
  *
- *	L C K _ g e t _ o w n e r _ h a n d l e
+ *	g e t _ o w n e r _ h a n d l e
  *
  **************************************
  *
@@ -489,24 +512,25 @@ SLONG LCK_get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 	{
 	case LCK_database:
 	case LCK_bdb:
+	case LCK_shadow:
+	case LCK_backup_alloc:
+	case LCK_backup_database:
+	case LCK_shared_counter:
+	case LCK_sweep:
+	case LCK_crypt:
+	case LCK_crypt_status:
+		handle = *LCK_OWNER_HANDLE_DBB(tdbb);
+		break;
+
+	case LCK_attachment:
 	case LCK_rel_exist:
 	case LCK_rel_partners:
 	case LCK_rel_rescan:
 	case LCK_idx_exist:
-	case LCK_shadow:
-	case LCK_retaining:
 	case LCK_expression:
 	case LCK_prc_exist:
-	case LCK_backup_alloc:
-	case LCK_backup_database:
-	case LCK_monitor:
+	case LCK_fun_exist:
 	case LCK_tt_exist:
-	case LCK_shared_counter:
-	case LCK_sweep:
-	case LCK_rel_gc:
-		handle = *LCK_OWNER_HANDLE_DBB(tdbb);
-		break;
-	case LCK_attachment:
 	case LCK_page_space:
 	case LCK_relation:
 	case LCK_tra:
@@ -515,11 +539,15 @@ SLONG LCK_get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 	case LCK_dsql_cache:
 	case LCK_backup_end:
 	case LCK_cancel:
+	case LCK_monitor:
 	case LCK_btr_dont_gc:
+	case LCK_rel_gc:
+	case LCK_record_gc:
 		handle = *LCK_OWNER_HANDLE_ATT(tdbb);
 		break;
+
 	default:
-		bug_lck("Invalid lock type in LCK_get_owner_handle()");
+		bug_lck("Invalid lock type in get_owner_handle()");
 	}
 
 	if (!handle)
@@ -531,33 +559,6 @@ SLONG LCK_get_owner_handle(thread_db* tdbb, enum lck_t lock_type)
 }
 
 
-SLONG LCK_get_owner_handle_by_type(thread_db* tdbb, lck_owner_t lck_owner_type)
-{
-/**********************************************************
- *
- *	L C K _ g e t _ o w n e r _ h a n d l e _ b y _ t y p e
- *
- **********************************************************
- *
- * Functional description
- *	return the lock owner given a lock owner type.
- *
- *********************************************************/
-	SET_TDBB(tdbb);
-
-	switch (lck_owner_type)
-	{
-		case LCK_OWNER_database:
-			return *LCK_OWNER_HANDLE_DBB(tdbb);
-		case LCK_OWNER_attachment:
-			return *LCK_OWNER_HANDLE_ATT(tdbb);
-		default:
-			bug_lck("Invalid lock owner type in LCK_get_owner_handle_by_type ()");
-			return 0;
-	}
-}
-
-
 void LCK_init(thread_db* tdbb, enum lck_owner_t owner_type)
 {
 /**************************************
@@ -593,14 +594,17 @@ void LCK_init(thread_db* tdbb, enum lck_owner_t owner_type)
 		break;
 	}
 
-	if (!dbb->dbb_lock_mgr->initializeOwner(tdbb, owner_id, owner_type, owner_handle_ptr))
+	FbLocalStatus statusVector;
+
+	if (!dbb->dbb_lock_mgr->initializeOwner(&statusVector, owner_id, owner_type, owner_handle_ptr))
 	{
-		if (tdbb->tdbb_status_vector[1] == isc_lockmanerr)
+		if (statusVector[1] == isc_lockmanerr)
 		{
+			statusVector.copyTo(tdbb->tdbb_status_vector);
 			tdbb->getDatabase()->dbb_flags |= DBB_bugcheck;
 		}
 
-		ERR_punt();
+		statusVector.raise();
 	}
 }
 
@@ -620,30 +624,42 @@ bool LCK_lock(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
 	SET_TDBB(tdbb);
 	fb_assert(LCK_CHECK_LOCK(lock));
 
+#ifdef DEBUG_LCK
+	LckSync sync(lock, "LCK_lock");
+#endif
+
 	Database* dbb = lock->lck_dbb;
-    set_lock_attachment(lock, tdbb->getAttachment());
+    lock->setLockAttachment(tdbb, tdbb->getAttachment());
 
 	WaitCancelGuard guard(tdbb, lock, wait);
-	ENQUEUE(tdbb, lock, level, wait);
+	FbLocalStatus statusVector;
+
+	ENQUEUE(tdbb, &statusVector, lock, level, wait);
 	fb_assert(LCK_CHECK_LOCK(lock));
+
 	if (!lock->lck_id)
 	{
-    	set_lock_attachment(lock, NULL);
+    	lock->setLockAttachment(tdbb, NULL);
 		if (!wait)
+		{
+			statusVector.copyTo(tdbb->tdbb_status_vector);
 			return false;
+		}
 
-		switch (tdbb->tdbb_status_vector[1])
+		switch (statusVector[1])
 		{
 		case isc_deadlock:
 		case isc_lock_conflict:
 		case isc_lock_timeout:
+			statusVector.copyTo(tdbb->tdbb_status_vector);
 			tdbb->checkCancelState(true);
 			return false;
 		case isc_lockmanerr:
 			dbb->dbb_flags |= DBB_bugcheck;
 			break;
 		}
-		ERR_punt();
+
+		statusVector.raise();
 	}
 
 	if (!lock->lck_compatible)
@@ -683,7 +699,7 @@ bool LCK_lock_opt(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
 }
 
 
-SLONG LCK_query_data(thread_db* tdbb, Lock* parent, enum lck_t lock_type, USHORT aggregate)
+SINT64 LCK_query_data(thread_db* tdbb, enum lck_t lock_type, USHORT aggregate)
 {
 /**************************************
  *
@@ -700,13 +716,11 @@ SLONG LCK_query_data(thread_db* tdbb, Lock* parent, enum lck_t lock_type, USHORT
 	SET_TDBB(tdbb);
 	Database* const dbb = tdbb->getDatabase();
 
-	fb_assert(LCK_CHECK_LOCK(parent));
-
-	return dbb->dbb_lock_mgr->queryData(parent->lck_id, lock_type, aggregate);
+	return dbb->dbb_lock_mgr->queryData(lock_type, aggregate);
 }
 
 
-SLONG LCK_read_data(thread_db* tdbb, Lock* lock)
+SINT64 LCK_read_data(thread_db* tdbb, Lock* lock)
 {
 /**************************************
  *
@@ -723,21 +737,10 @@ SLONG LCK_read_data(thread_db* tdbb, Lock* lock)
 
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-#ifdef VMS
-	if (!LCK_lock(NULL, lock, LCK_null, LCK_NO_WAIT))
-		return 0;
-
-	const SLONG data = dbb->dbb_lock_mgr->readData(lock->lck_id);
-	LCK_release(lock);
-#else
-	Lock* parent = lock->lck_parent;
-	const SLONG data =
-		dbb->dbb_lock_mgr->readData2(parent ? parent->lck_id : 0,
-									 lock->lck_type,
-									 (UCHAR*) &lock->lck_key, lock->lck_length,
+	const SINT64 data =
+		dbb->dbb_lock_mgr->readData2(lock->lck_type,
+									 lock->getKeyString(), lock->lck_length,
 									 lock->lck_owner_handle);
-#endif
-
 	fb_assert(LCK_CHECK_LOCK(lock));
 	return data;
 }
@@ -758,13 +761,17 @@ void LCK_release(thread_db* tdbb, Lock* lock)
 	SET_TDBB(tdbb);
 	fb_assert(LCK_CHECK_LOCK(lock));
 
+#ifdef DEBUG_LCK
+	LckSync sync(lock, "LCK_release");
+#endif
+
 	if (lock->lck_physical != LCK_none) {
 		DEQUEUE(tdbb, lock);
 	}
 
 	lock->lck_physical = lock->lck_logical = LCK_none;
 	lock->lck_id = lock->lck_data = 0;
-	set_lock_attachment(lock, NULL);
+	lock->setLockAttachment(tdbb, NULL);
 
 	fb_assert(LCK_CHECK_LOCK(lock));
 }
@@ -802,7 +809,7 @@ void LCK_re_post(thread_db* tdbb, Lock* lock)
 }
 
 
-void LCK_write_data(thread_db* tdbb, Lock* lock, SLONG data)
+void LCK_write_data(thread_db* tdbb, Lock* lock, SINT64 data)
 {
 /**************************************
  *
@@ -884,7 +891,7 @@ static bool compatible(const Lock* lock1, const Lock* lock2, USHORT level2)
 }
 
 
-static void enqueue(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
+static void enqueue(thread_db* tdbb, CheckStatusWrapper* statusVector, Lock* lock, USHORT level, SSHORT wait)
 {
 /**************************************
  *
@@ -901,19 +908,10 @@ static void enqueue(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
 
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-	Lock* parent = lock->lck_parent;
-	lock->lck_id = dbb->dbb_lock_mgr->enqueue(tdbb,
-											  lock->lck_id,
-											  parent ? parent->lck_id : 0,
-											  lock->lck_type,
-											  (const UCHAR*) &lock->lck_key,
-											  lock->lck_length,
-											  level,
-											  lock->lck_ast,
-											  lock->lck_object,
-											  lock->lck_data,
-											  wait,
-											  lock->lck_owner_handle);
+	lock->lck_id = dbb->dbb_lock_mgr->enqueue(tdbb, statusVector, lock->lck_id,
+		lock->lck_type, lock->getKeyString(), lock->lck_length,
+		level, lock->lck_ast, lock->lck_object, lock->lck_data, wait,
+		lock->lck_owner_handle);
 
 	if (!lock->lck_id)
 	{
@@ -924,7 +922,7 @@ static void enqueue(thread_db* tdbb, Lock* lock, USHORT level, SSHORT wait)
 }
 
 
-static int external_ast(void* ast_object)
+static int external_ast(void* lock_void)
 {
 /**************************************
  *
@@ -938,29 +936,21 @@ static int external_ast(void* ast_object)
  *	we are blocking a lock from another process.
  *
  **************************************/
-	Lock* const lock = static_cast<Lock*>(ast_object);
+	Lock* lock = static_cast<Lock*>(lock_void);
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-	try
+	// go through the list, saving the next lock in the list
+	// in case the current one gets deleted in the ast
+
+	Lock* next;
+	for (Lock* match = hash_get_lock(lock, 0, 0); match; match = next)
 	{
-		AstContextHolder tdbb(lock->lck_dbb, lock->lck_attachment);
-
-		// go through the list, saving the next lock in the list
-		// in case the current one gets deleted in the ast
-
-		Lock* next;
-		for (Lock* match = hash_get_lock(lock, 0, 0); match; match = next)
-		{
-			next = match->lck_identical;
-			if (match->lck_ast) {
-				(*match->lck_ast)(match->lck_object);
-			}
+		next = match->lck_identical;
+		if (match->lck_ast) {
+			(*match->lck_ast)(match->lck_object);
 		}
 	}
-	catch (const Firebird::Exception&)
-	{} // no-op
-
-	return 0;
+	return 0; // make the compiler happy
 }
 
 
@@ -989,7 +979,7 @@ static USHORT hash_func(const UCHAR* value, USHORT length)
 	{
 		if (!(l & 3))
 			p = (UCHAR*) &hash_value;
-		*p++ = *q++;
+		*p++ += *q++;
 	}
 
 	return (USHORT) (hash_value % LOCK_HASH_SIZE);
@@ -1011,9 +1001,8 @@ static void hash_allocate(Lock* lock)
  **************************************/
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-	Database* dbb = lock->lck_dbb;
+	Jrd::Attachment* const attachment = lock->getLockAttachment();
 
-	Attachment* attachment = lock->lck_attachment;
 	if (attachment)
 	{
 		attachment->att_compatibility_table =
@@ -1040,14 +1029,14 @@ static Lock* hash_get_lock(Lock* lock, USHORT* hash_slot, Lock*** prior)
  **************************************/
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-	Attachment* const att = lock->lck_attachment;
+	Jrd::Attachment* const att = lock->getLockAttachment();
 	if (!att)
 		return NULL;
 
 	if (!att->att_compatibility_table)
 		hash_allocate(lock);
 
-	const USHORT hash_value = hash_func((UCHAR*) & lock->lck_key, lock->lck_length);
+	const USHORT hash_value = hash_func(lock->getKeyString(), lock->lck_length);
 
 	if (hash_slot)
 		*hash_slot = hash_value;
@@ -1067,14 +1056,12 @@ static Lock* hash_get_lock(Lock* lock, USHORT* hash_slot, Lock*** prior)
 	for (Lock* collision = match; collision; collision = collision->lck_collision)
 	{
 		fb_assert(LCK_CHECK_LOCK(collision));
-		if (collision->lck_parent && lock->lck_parent &&
-			collision->lck_parent->lck_id == lock->lck_parent->lck_id &&
-			collision->lck_type == lock->lck_type &&
+		if (collision->lck_type == lock->lck_type &&
 			collision->lck_length == lock->lck_length)
 		{
 			// check that the keys are the same
 
-			if (!memcmp(lock->lck_key.lck_string, collision->lck_key.lck_string, lock->lck_length))
+			if (!memcmp(lock->getKeyString(), collision->getKeyString(), lock->lck_length))
 				return collision;
 		}
 
@@ -1101,7 +1088,7 @@ static void hash_insert_lock(Lock* lock)
  **************************************/
 	fb_assert(LCK_CHECK_LOCK(lock));
 
-	Attachment* const att = lock->lck_attachment;
+	Jrd::Attachment* const att = lock->getLockAttachment();
 	if (!att)
 		return;
 
@@ -1306,11 +1293,13 @@ static void internal_dequeue(thread_db* tdbb, Lock* lock)
 
 	// check for a potential downgrade
 
-	internal_downgrade(tdbb, match);
+	FbLocalStatus statusVector;
+	internal_downgrade(tdbb, &statusVector, match);
+	fb_assert(statusVector.isEmpty());
 }
 
 
-static USHORT internal_downgrade(thread_db* tdbb, Lock* first)
+static USHORT internal_downgrade(thread_db* tdbb, CheckStatusWrapper* statusVector, Lock* first)
 {
 /**************************************
  *
@@ -1340,7 +1329,8 @@ static USHORT internal_downgrade(thread_db* tdbb, Lock* first)
 
 	if (level < first->lck_physical)
 	{
-		if (dbb->dbb_lock_mgr->convert(tdbb, first->lck_id, level, LCK_NO_WAIT, external_ast, first))
+		if (dbb->dbb_lock_mgr->convert(tdbb, statusVector, first->lck_id, level, LCK_NO_WAIT,
+				external_ast, first))
 		{
 			for (Lock* lock = first; lock; lock = lock->lck_identical)
 			{
@@ -1355,10 +1345,8 @@ static USHORT internal_downgrade(thread_db* tdbb, Lock* first)
 }
 
 
-static bool internal_enqueue(thread_db* tdbb,
-							Lock* lock,
-							USHORT level,
-							SSHORT wait, bool convert_flg)
+static bool internal_enqueue(thread_db* tdbb, CheckStatusWrapper* statusVector, Lock* lock,
+	USHORT level, SSHORT wait, bool convert_flg)
 {
 /**************************************
  *
@@ -1382,8 +1370,6 @@ static bool internal_enqueue(thread_db* tdbb,
 	fb_assert(LCK_CHECK_LOCK(lock));
 	fb_assert(lock->lck_compatible);
 
-	ISC_STATUS* const status = tdbb->tdbb_status_vector;
-
 	// look for an identical lock
 
 	Lock* match = hash_get_lock(lock, 0, 0);
@@ -1396,7 +1382,7 @@ static bool internal_enqueue(thread_db* tdbb,
 			// for now return a lock conflict; it would be better if we were to
 			// do a wait on the other lock by setting some flag bit or some such
 
-			Arg::Gds(isc_lock_conflict).copyTo(status);
+			(Arg::StatusVector(statusVector) << Arg::Gds(isc_lock_conflict)).copyTo(statusVector);
 			return false;
 		}
 
@@ -1410,12 +1396,8 @@ static bool internal_enqueue(thread_db* tdbb,
 
 			if (level > match->lck_physical)
 			{
-				if (!dbb->dbb_lock_mgr->convert(tdbb,
-												match->lck_id,
-												level,
-												wait,
-												external_ast,
-												lock))
+				if (!dbb->dbb_lock_mgr->convert(tdbb, statusVector, match->lck_id, level, wait,
+						external_ast, lock))
 				{
 					return false;
 				}
@@ -1443,18 +1425,9 @@ static bool internal_enqueue(thread_db* tdbb,
 	// enqueue the lock, but swap out the ast and the ast argument
 	// with the local ast handler, passing it the lock block itself
 
-	lock->lck_id = dbb->dbb_lock_mgr->enqueue(tdbb,
-											  lock->lck_id,
-											  lock->lck_parent ? lock->lck_parent->lck_id : 0,
-											  lock->lck_type,
-											  (const UCHAR*) &lock->lck_key,
-											  lock->lck_length,
-											  level,
-											  external_ast,
-											  lock,
-											  lock->lck_data,
-											  wait,
-											  lock->lck_owner_handle);
+	lock->lck_id = dbb->dbb_lock_mgr->enqueue(tdbb, statusVector, lock->lck_id,
+		lock->lck_type, lock->getKeyString(), lock->lck_length,
+		level, external_ast, lock, lock->lck_data, wait, lock->lck_owner_handle);
 
 	// If the lock exchange failed, set the lock levels appropriately
 	if (lock->lck_id == 0)
@@ -1475,60 +1448,99 @@ static bool internal_enqueue(thread_db* tdbb,
 	return lock->lck_id ? true : false;
 }
 
-
-static void set_lock_attachment(Lock* lock, Attachment* attachment)
+Lock::Lock(thread_db* tdbb, USHORT length, lck_t type, void* object, lock_ast_t ast)
+:	lck_dbb(tdbb->getDatabase()),
+ 	lck_attachment(NULL),
+	lck_compatible(NULL),
+	lck_compatible2(NULL),
+	lck_ast(ast),
+	lck_object(object),
+	lck_next(NULL),
+	lck_prior(NULL),
+	lck_collision(NULL),
+	lck_identical(NULL),
+	lck_id(0),
+	lck_owner_handle(get_owner_handle(tdbb, type)),
+	lck_length(length),
+	lck_type(type),
+	lck_logical(LCK_none),
+	lck_physical(LCK_none),
+	lck_data(0)
 {
-	if (lock->lck_attachment == attachment)
+	lck_key.lck_long = 0;
+	lck_tail[0] = 0;
+}
+
+void Lock::setLockAttachment(thread_db* tdbb, Jrd::Attachment* attachment)
+{
+	SET_TDBB(tdbb);
+	Database* dbb = tdbb->getDatabase();
+	fb_assert(dbb);
+	if (!dbb)
+		return;
+
+	Sync lckSync(&dbb->dbb_lck_sync, "setLockAttachment");
+	lckSync.lock(SYNC_EXCLUSIVE);
+
+	Attachment* att = lck_attachment ? lck_attachment->getHandle() : NULL;
+	if (att == attachment)
 		return;
 
 	// If lock has an attachment it must not be a part of linked list
-	fb_assert(!lock->lck_attachment ? !lock->lck_prior && !lock->lck_next : true);
+	fb_assert(!lck_attachment ? !lck_prior && !lck_next : true);
 
 	// Delist in old attachment
-	if (lock->lck_attachment)
+	if (att)
 	{
 		// Check that attachment seems to be valid, check works only when DEBUG_GDS_ALLOC is defined
-		fb_assert(lock->lck_attachment->att_flags != 0xDEADBEEF);
+		fb_assert(att->att_flags != 0xEEEEEEEE);
 
-		Lock* const next = lock->lck_next;
-		Lock* const prior = lock->lck_prior;
-
-		if (prior)
+		if (lck_prior)
 		{
-			fb_assert(prior->lck_next == lock);
-			prior->lck_next = next;
+			fb_assert(lck_prior->lck_next == this);
+			lck_prior->lck_next = lck_next;
 		}
 		else
 		{
-			fb_assert(lock->lck_attachment->att_long_locks == lock);
-			lock->lck_attachment->att_long_locks = next;
+			fb_assert(att->att_long_locks == this);
+			att->att_long_locks = lck_next;
 		}
 
-		if (next)
+		if (lck_next)
 		{
-			fb_assert(next->lck_prior == lock);
-			next->lck_prior = prior;
+			fb_assert(lck_next->lck_prior == this);
+			lck_next->lck_prior = lck_prior;
 		}
 
-		lock->lck_next = NULL;
-		lock->lck_prior = NULL;
+		lck_next = NULL;
+		lck_prior = NULL;
 	}
 
-
 	// Enlist in new attachment
 	if (attachment)
 	{
 		// Check that attachment seems to be valid, check works only when DEBUG_GDS_ALLOC is defined
-		fb_assert(attachment->att_flags != 0xDEADBEEF);
+		fb_assert(attachment->att_flags != 0xEEEEEEEE);
 
-		lock->lck_next = attachment->att_long_locks;
-		lock->lck_prior = NULL;
-		attachment->att_long_locks = lock;
+		lck_next = attachment->att_long_locks;
+		lck_prior = NULL;
+		attachment->att_long_locks = this;
 
-		Lock* next = lock->lck_next;
-		if (next)
-			next->lck_prior = lock;
+		if (lck_next)
+			lck_next->lck_prior = this;
 	}
 
-	lock->lck_attachment = attachment;
+	RefDeb(DEB_RLS_JATT, "setLockAttachment");
+	lck_attachment = attachment ? attachment->getStable() : NULL;
+}
+
+Lock* Lock::detach()
+{
+	Lock* next = lck_next;
+	RefDeb(DEB_RLS_JATT, "Lock::detach");
+	lck_attachment = NULL;
+	lck_next = NULL;
+	lck_prior = NULL;
+
+	return next;
 }
